/*
 *  ps_read_pdb_file.c
 *  ProteinSplitter
 *
 *  Created by Naoto Morikawa on 7/21/06.
 *  Copyright (C) 2006 Naoto Morikawa. All rights reserved.
 *
 *  2006-10-01 version 0.1 release
 */

/**************************************/
/*  Second scan for data collection   */
/**************************************/
/*   atm        seq           mod     */
/*          atom_head      seq_head   */
/*          atom_tail      seq_tail   */
/*           (id)                      */
/*           (mod_id)                 */
/*           (mol_id)                  */
/**************************************/
/*
 1 <= cur_mod <= my_protein.mod_num
 my_protein.atom[my_protein.sequence[my_protein.model[cur_mod].seq_head].atom_head]
 my_protein.atom[my_protein.sequence[my_protein.model[cur_mod].seq_head].atom_tail]
 ....
 my_protein.atom[my_protein.sequence[my_protein.model[cur_mod].seq_tail].atom_head]
 my_protein.atom[my_protein.sequence[my_protein.model[cur_mod].seq_tail].atom_tail]
 
 1 <= cur_seq <= my_protein.seq_num
 my_protein.atom[my_protein.sequence[cur_seq].atom_head]
 my_protein.atom[my_protein.sequence[cur_seq].atom_tail]
 */
/*
 for (p_ct->model=1; p_ct->model <= my_protein.mod_num; p_ct->model++) {
	 for ( p_ct->sequence=my_protein.model[p_ct->model].seq_head; p_ct->sequence <=my_protein.model[p_ct->model].seq_tail; p_ct->sequence++) {
		 for (p_ct->sequence=my_protein.sequence[p_ct->sequence].atom_head; p_ct->sequence <=my_protein.sequence[p_ct->sequence].atom_tail; p_ct->sequence++) {
		 }
	 }
 }
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "ps_main.h"
#include "ps_pdb_data.h"
#include "ps_print.h"

struct pe_pdb_header my_pdb_header;
struct pe_pdb_master my_pdb_master;
extern struct pe_protein my_protein;

int atom_conversion(struct pe_atom * pt_atm, struct pe_pdb_atom * pt_pdb_atm)
{
	char work4[4];
	char work5[5];
	char work6[6];
	char work8[8];

	strncpy(work5, pt_pdb_atm->serial_num, 5);
	pt_atm->serial_num = atol(work5);
	
	if (!strncmp(" CA ", pt_pdb_atm->atom_name, 4)) {
		pt_atm->atom_name = PE_CA;
	} else if (!strncmp(" C  ", pt_pdb_atm->atom_name, 4)) {
		pt_atm->atom_name = PE_CC;
	} else if (!strncmp(" N  ", pt_pdb_atm->atom_name, 4)) {
		pt_atm->atom_name = PE_NN;
	} else {
		fprintf(STDERR, "atom_comversion() : ERROR  unknown atom\n");
		print_pdb_atom(pt_pdb_atm);
		print_my_atom(pt_atm);
		return FAIL;
	}
	pt_atm->alt_loc = pt_pdb_atm->alt_loc;
	strncpy(pt_atm->res_name, pt_pdb_atm->res_name, 3);
	pt_atm->chain_id = pt_pdb_atm->chain_id;
	
	strncpy(work4, pt_pdb_atm->res_seq, 4);
	pt_atm->res_seq = atol(work4);
	
	strncpy(work8, pt_pdb_atm->x, 8);
	pt_atm->x = atof(work8);
	
	strncpy(work8, pt_pdb_atm->y, 8);
	pt_atm->y = atof(work8);
	
	strncpy(work8, pt_pdb_atm->z, 8);
	pt_atm->z = atof(work8);
	
	strncpy(work6, pt_pdb_atm->occupancy, 6);
	pt_atm->occupancy = atof(work6);
	
	strncpy(pt_atm->seg_id, pt_pdb_atm->seg_id, 4);
	return SUCCESS;
}

int extract_molecule_id(char * string, int cur_ct)
{
	char * ptr;
	char tmp[3];

	ptr = string;
	while (*ptr == ' ') ptr++;
	cur_ct++;
	if (cur_ct > PDB_MOLECULE_MAX) {
		fprintf(STDERR,
	"extract_molecule_id() : ERROR  - molecule[%d] OVERFLOW - \n",
			cur_ct);
		return FAIL;;
	}
	
	tmp[0] = *ptr++;
	if ((*ptr != ' ') && (*ptr != ';')) {	/* double digits? */
		tmp[1] = *ptr++;
		tmp[2] = '\0';
	} else {
		tmp[1] = '\0';
	}
	while (*ptr == ' ')
		ptr++;
	if (*ptr != ';') {
		fprintf(STDERR,
	"extract_molecule_id() : ERROR  - molecule[%d].id format error - \n",
			cur_ct);
		return FAIL;
	}

	my_protein.molecule[cur_ct].id = atoi(tmp);
	return SUCCESS;
}

int extract_chain_id(char * string, int cur_ct)
{
	char * ptr;
	int flag;

	ptr = string;
	flag = 0;
	while (*ptr == ' ') ptr++;
	while (*ptr != NULL) {
		cur_ct++;
		if (cur_ct > PDB_CHAIN_MAX) {
			fprintf(STDERR,
			"extract_chain_id() : ERROR  - chain[%d] OVERFLOW - \n",
				cur_ct);
			exit(1);	/* [FIXME] */
		}
		my_protein.chain[cur_ct].id = *ptr++;
		if (*ptr == 'U') {
			/* the 2nd char of 'NULL' */
			my_protein.chain[cur_ct].id = ' ';
			ptr = ptr + 3;
		} else {
			while ((*ptr != ' ') && (*ptr != ';')
			       && (*ptr != ',')) {
				/* Set the last char */
				my_protein.chain[cur_ct].id = *ptr++;
			}
		}
		while ((*ptr == ' '))
			ptr++;
		while (1) {
			if (*ptr == ';')
				return cur_ct;
			if (*ptr == ',') {
				flag = 1;
				break;
			}
			fprintf(STDERR,
			"extract_chain_id() : format ERROR (missing ';') \n");
			return cur_ct;	/* the case of missing ';' */
		}
		ptr++;
		while ((*ptr == ' '))
			ptr++;
	}
	if (flag > 0) {
		return -1 * cur_ct;	/* fetch the next line */
	} else {
		fprintf(STDERR, "extract_chain_id() : format ERROR \n");
		exit(1);	/* [FIXME] */
	}
	/* NOTREACHED */
}

void initialize_protein_data_arrays(void)
{
	int i;

	my_protein.mod_num = 0;
	my_protein.model = NULL;
	my_protein.mol_num = 0;
	my_protein.chn_num = 0;
	my_protein.res_num = 0;
	my_protein.atom = NULL;
	my_protein.seq_num = 0;
	my_protein.sequence = NULL;
	my_protein.tile_max_num = 0;
	my_protein.tile_num = 0;
	
	/* [Note] 0-th entry of the arrays will not be used. */
	for (i = 0; i < PDB_MOLECULE_MAX + 1; i++) {
		my_protein.molecule[i].chain_head = 0;
		my_protein.molecule[i].chain_tail = 0;
		my_protein.molecule[i].id = 0;
		strcpy(my_protein.molecule[i].name, "\0");
	}

	for (i = 0; i < PDB_CHAIN_MAX + 1; i++) {
		my_protein.chain[i].atom_head = 0;
		my_protein.chain[i].atom_tail = 0;
		my_protein.chain[i].id = '-';
		my_protein.chain[i].mol_id = 0;
		my_protein.chain[i].mod_num = 0;
		my_protein.chain[i].tile_head = 0;
		my_protein.chain[i].tile_tail = 0;
	}
	return;
}

struct pe_counter {
	int model;
	int molecule;
	int chain;
	int residue;
	int sequence;
	int atom;
};

void print_counter(struct pe_counter * p_ct)
{
	fprintf(STDOUT, "\n");
	fprintf(STDOUT, "[counter] model    = %d\n", p_ct->model);
	fprintf(STDOUT, "[counter] molecule = %d\n", p_ct->molecule);
	fprintf(STDOUT, "[counter] chain    = %d\n", p_ct->chain);
	fprintf(STDOUT, "[counter] residue  = %d\n", p_ct->residue);
	fprintf(STDOUT, "[counter] sequence = %d\n", p_ct->sequence);
	fprintf(STDOUT, "[counter] atom     = %d\n", p_ct->atom);
	return;
}

void count_pdb_atom(char * pdb_line, struct pe_counter * p_ct)
{
	/*
	 * [FIXME] Alternate location indicaters in ATOM are assumed to be
	 * A, B, C, ...
	 */
	if (!strncmp("ATOM", pdb_line, 4)) {
		if (0)
			print_pdb_atom((struct pe_pdb_atom *) pdb_line);
		if (!strncmp("CA", pdb_line + 13, 2)
		    && ((pdb_line[16] == ' ')
			|| (pdb_line[16] == 'A')))
			p_ct->residue++;
	}
	return;
}

int get_chain_identifier(FILE * fp_pdb, char * pdb_line,
			 struct pe_counter * p_ct)
{
	/* The chain count was incremented when the molecule's was. */
	p_ct->chain--;	
	my_protein.molecule[p_ct->molecule].chain_head = p_ct->chain + 1;
	p_ct->chain = extract_chain_id(pdb_line + 17, p_ct->chain);
	while (p_ct->chain < 0) {
		p_ct->chain = -1 * p_ct->chain;
		fgets(pdb_line, LINE_WIDTH + 1, fp_pdb);
		while (*pdb_line != 'C') {
			if (fgets(pdb_line, LINE_WIDTH + 1, fp_pdb) == NULL) {
				fprintf(STDERR, 
"get_chain_identifier() : format ERROR  chain[%d] OVERFLOW - \n",
					p_ct->chain);
				return FAIL;
			}
		}
		if (!strncmp("COMP", pdb_line, 4)) {
			p_ct->chain = extract_chain_id(pdb_line + 11, 
						       p_ct->chain);
		} else {
			print_pdb_line((struct pe_pdb_line *) pdb_line);
			fprintf(STDERR,
"get_chain_identifier() : format ERROR  chain[%d] OVERFLOW - \n",
				p_ct->chain);
			return FAIL;
		}
	}
	my_protein.molecule[p_ct->molecule].chain_tail = p_ct->chain;
	return SUCCESS;
}

int count_pdb_c_line(FILE * fp_pdb, char * pdb_line, struct pe_counter * p_ct)
{
	char * l_ptr;
	int wk_ct;
	
	if (!strncmp("COMP", pdb_line, 4)) {
		if (DEBUG)
			print_pdb_compnd((struct pe_pdb_compnd *) pdb_line);
		if ((pdb_line[9] == ' ')
		    && !strncmp(pdb_line + 10, "MOL_ID: ", 8)) {
			/* get molecule identifier */
			if (extract_molecule_id(pdb_line + 17, p_ct->molecule)
			    == FAIL)
				return FAIL;
			p_ct->molecule++;
			p_ct->chain++;
			my_protein.chain[p_ct->chain].id = ' ';
			my_protein.molecule[p_ct->molecule].chain_head
				= p_ct->chain;
			my_protein.molecule[p_ct->molecule].chain_tail
				= p_ct->chain;
		} else if (!strncmp(pdb_line + 11, "MOL_ID: ", 8)) {
			/* get molecule identifier */
			if (extract_molecule_id(pdb_line + 18, p_ct->molecule)
			    == FAIL)
				return FAIL;
			p_ct->molecule++;
			p_ct->chain++;
			my_protein.chain[p_ct->chain].id = ' ';
			my_protein.molecule[p_ct->molecule].chain_head
				= p_ct->chain;
			my_protein.molecule[p_ct->molecule].chain_tail
				= p_ct->chain;
		} else if (!strncmp(pdb_line + 11, "MOLECULE: ",10)) {
			/* get molecule name */
			l_ptr = pdb_line + 20;
			wk_ct = 0;
			while (*l_ptr == ' ') l_ptr++;
			while ((*l_ptr != ';') && (*l_ptr != ',')
			       && (*l_ptr != '\n') && (wk_ct < 70)) {
				my_protein.molecule[p_ct->molecule].
				name[wk_ct++] = *l_ptr++;
			}
			my_protein.molecule[p_ct->molecule].name[wk_ct] = '\0';
		} else if (!strncmp(pdb_line + 11, "CHAIN: ", 7)) {
			if (get_chain_identifier(fp_pdb, pdb_line, p_ct)
			    == FAIL)
				return FAIL;
		}
	}
	return SUCCESS;
}

int count_pdb_e_line(char * pdb_line, struct pe_counter * p_ct)
{
	if (!strncmp("ENDM", pdb_line, 4)) {
		/* The end of the NMR model */
		p_ct->model++;
	} else if (!strncmp("END", pdb_line, 3)) {
		/* Do nothing. This is the end of the PDB file */
	}
	return SUCCESS;
}

int count_pdb_h_line(char * pdb_line, struct pe_counter * p_ct)
{
	if (!strncmp("HEAD", pdb_line, 4)) {
		if (0) print_pdb_header((struct pe_pdb_header *) pdb_line);
		strcpy((char *)&my_pdb_header, pdb_line);
	}
	return SUCCESS;
}

int count_pdb_m_line(char * pdb_line, struct pe_counter * p_ct)
{
	if (!strncmp("MODE", pdb_line, 4)) {
		/* The start of new NMR model */
		if (0) print_pdb_model((struct pe_pdb_model *) pdb_line);
	} else if (!strncmp("MAST", pdb_line, 4)) {
		strcpy((char *)&my_pdb_master, pdb_line);
		if (0) print_pdb_master(&my_pdb_master);
	}
	return SUCCESS;
}

int count_pdb_t_line(char * pdb_line, struct pe_counter * p_ct)
{
	if (!strncmp("TER", pdb_line, 3)) {
		/* The end of a chain */
		if (0) print_pdb_ter((struct pe_pdb_ter *) pdb_line);
		p_ct->sequence++;
	}
	return SUCCESS;
}

int count_pdb_general(FILE * fp_pdb, char * pdb_line, struct pe_counter * p_ct)
{
	switch (*pdb_line) {
		case ('C'):
			if (count_pdb_c_line(fp_pdb, pdb_line, p_ct) == FAIL)
				return FAIL;
			break;
		case ('E'):
			if (count_pdb_e_line(pdb_line, p_ct) == FAIL)
				return FAIL;
			break;
		case ('H'):
			if (count_pdb_h_line(pdb_line, p_ct) == FAIL)
				return FAIL;
			break;
		case ('M'):
			if (count_pdb_m_line(pdb_line, p_ct) == FAIL)
				return FAIL;
			break;
		case ('T'):
			if (count_pdb_t_line(pdb_line, p_ct) == FAIL)
				return FAIL;
			break;
	}
	return SUCCESS;
}

int count_pdb_records(FILE * fp_pdb, struct pe_counter * p_ct)
{
	char pdb_line[LINE_WIDTH + 1];
	
	p_ct->model = 0;
	p_ct->molecule = 0;
	p_ct->chain = 0;
	p_ct->residue = 0;
	p_ct->sequence = 0;
	p_ct->atom = 0;

	while (fgets(pdb_line, LINE_WIDTH + 1, fp_pdb) != NULL) {
		if (0) {
			fprintf(STDOUT, 
				"count_pdb_records(): pdb_line = %s\n",
				pdb_line);
		}
		if (*pdb_line == 'A') {
			count_pdb_atom(pdb_line, p_ct);
		} else {
			if (count_pdb_general(fp_pdb, pdb_line, p_ct) == FAIL)
				return FAIL;

			/* [FIXME] Only the first NMR model is analyzed. */
			if (p_ct->model >= 1) {
				if (p_ct->residue == 0) {
					fprintf(STDERR,
"count_pdb_records(): ERROR  - No CA atoms (DNA or RNA molecules only)  \n");
					return FAIL;
				} else {
					return SUCCESS;
				}
			}
		}
	}	/* while */
	return SUCCESS;
}

int allocate_protein_models(struct pe_counter * p_ct)
{
	struct pe_model * mdl;
	int model_size;
	int i;
	
	model_size = 1+ p_ct->model + 1;	/* 0-th + model + dummy */
	mdl = (struct pe_model *) malloc(model_size * sizeof(struct pe_model));
	if (mdl == NULL) {
		fprintf(STDERR, 
		"allocate_protein_models() : ERROR allocation failed\n");
		return FAIL;
	}
	for (i = 0; i < model_size; i++) {
		mdl[i].seq_head = 0;
		mdl[i].seq_tail = 0;
	}
	my_protein.model = mdl;
	return SUCCESS;
}

int allocate_protein_sequences(struct pe_counter * p_ct)
{
	struct pe_chain * seq;
	int seq_size;
	int i;
	
	seq_size = 1 + p_ct->sequence + 1;	/* 0-th + sequence + dummy */
	seq = (struct pe_chain *) malloc(seq_size * sizeof(struct pe_chain));
	if (seq == NULL) {
		fprintf(STDERR, 
		"allocate_protein_sequences() : ERROR allocation failed\n");
		return FAIL;
	}
	for (i = 0; i < seq_size; i++) {
		seq[i].atom_head = 0;
		seq[i].atom_tail = 0;
		seq[i].id = '-';
		seq[i].mol_id = 0;
		seq[i].mod_num = 0;
		seq[i].tile_head = 0;
		seq[i].tile_tail = 0;
	}
	my_protein.sequence = seq;
	return SUCCESS;
}

/* Set atom[loc] to DNA and RNA, i.e., no CA atoms. */
void set_RNA(int num, struct pe_atom * atm)
{
	atm[num].serial_num = 0;
	atm[num].atom_name = PE_RNA;	/* DNA or RNA */
	atm[num].alt_loc = ' ';
	strncpy(atm[num].res_name, "RNA", 3);
	atm[num].chain_id = ' ';
	atm[num].res_seq = 0;
	atm[num].x = 0.0;
	atm[num].y = 0.0;
	atm[num].z = 0.0;
	atm[num].occupancy = 0.0;
	strncpy(atm[num].seg_id, "    ", 4);
	atm[num].err = 0.0;
	atm[num].can_be_init_atom = YES;
	return;
}

int allocate_protein_atoms(struct pe_counter * p_ct)
{
	struct pe_atom * atm;
	int atom_size;
	int i;
	
	/*  Three entries per residue (for N, C, and CA) + 0-th */
	atom_size = 3 * p_ct->residue + 1;
	atm = (struct pe_atom *) malloc(atom_size * sizeof(struct pe_atom));
	if (atm == NULL) {
		fprintf(STDERR, 
		"allocate_protein_sequences() : ERROR allocation failed\n");
		return FAIL;
	}
	for (i = 0; i < atom_size; i++) {
		atm[i].atom_name = PE_UNKNOWN;	/* Mr. Nobody! */
		atm[i].can_be_init_atom = YES;
	}
	my_protein.atom = atm;
	
	set_RNA(0, my_protein.atom);
	return SUCCESS;
}

/* [Note] 0-th entry of the arrays will not be used. */
int allocate_protein_data_arrays(struct pe_counter * p_ct)
{
	int i;
	
	if (p_ct->model != 0)
		my_protein.mod_num = p_ct->model;
	else	/* Recall that the model record is optional in PDB files. */
		my_protein.mod_num = 1;

	if (p_ct->molecule != 0)
		my_protein.mol_num = p_ct->molecule;
	else {
		my_protein.mol_num = 1;
		strcpy(my_protein.molecule[1].name, "Not Available\0");
		my_protein.molecule[1].chain_head = 1;
		my_protein.molecule[1].chain_tail = p_ct->sequence;
	}
	
	if (p_ct->chain != 0)
		my_protein.chn_num = p_ct->chain;
	else {
		my_protein.chn_num = p_ct->sequence;
		for (i = 1; i <= p_ct->sequence; i++)
			my_protein.chain[i].mod_num = 1;
	}

	my_protein.res_num = p_ct->residue;	/* Count of CA atoms */

	if (p_ct->sequence > 0)
		my_protein.seq_num = p_ct->sequence;	/* count of TER */
	else {
		fprintf(STDERR, "read_pdb_file() : format ERROR (no TERs)\n");
		return FAIL;
	}
	
	if (allocate_protein_models(p_ct) == FAIL)
		return FAIL;

	if (allocate_protein_sequences(p_ct) == FAIL)
		return FAIL;

	if (allocate_protein_atoms(p_ct) == FAIL)
		return FAIL;
	return SUCCESS;
}

int read_pdb_atom(char * pdb_line, struct pe_counter * p_ct)
{
	int result = SUCCESS;
	
	if (!strncmp("ATOM", pdb_line, 4)) {
		if (0) print_pdb_atom((struct pe_pdb_atom *) pdb_line);
		if (!strncmp("N ", pdb_line + 13, 2)
		    && ((pdb_line[16] == ' ') || (pdb_line[16] == 'A')) ) {
			result = atom_conversion(my_protein.atom + p_ct->atom,
					(struct pe_pdb_atom *) pdb_line);
			if (result == FAIL) return FAIL;
			p_ct->atom++;
		} else if (!strncmp("CA", pdb_line + 13, 2)
			   && ((pdb_line[16] == ' ')
			       || (pdb_line[16] == 'A'))) {
			result = atom_conversion(my_protein.atom + p_ct->atom,
					(struct pe_pdb_atom *) pdb_line);
			if (result == FAIL) return FAIL;
			p_ct->atom++;
		} else if (!strncmp("C ", pdb_line + 13, 2)
			   && ((pdb_line[16] == ' ')
			       || (pdb_line[16] == 'A'))) {
			result = atom_conversion(my_protein.atom + p_ct->atom,
					(struct pe_pdb_atom *) pdb_line);
			if (result == FAIL) return FAIL;
			p_ct->atom++;
		}
	}
	return result;
}

void read_pdb_e_line(char * pdb_line, struct pe_counter * p_ct)
{
	if (!strncmp("ENDM", pdb_line, 4)) {
		/* The end of the NMR model */
		my_protein.model[p_ct->model].seq_tail = p_ct->sequence - 1;
		p_ct->model++;
	} else if (!strncmp("END", pdb_line, 3)) {
		/* The end of the PDB file */
	}
	return;
}

void read_pdb_m_line(char * pdb_line, struct pe_counter * p_ct)
{
	if (!strncmp("MODE", pdb_line, 4)) {
		/* The start of a new NMR model */
		if (0) print_pdb_model((struct pe_pdb_model *) pdb_line);
		p_ct->chain = 1;
		my_protein.model[p_ct->model].seq_head = p_ct->sequence;
		my_protein.sequence[p_ct->sequence].mod_num = p_ct->model;
		my_protein.sequence[p_ct->sequence].id 
			= my_protein.chain[p_ct->chain].id;
	}
	return;
}

void read_pdb_t_line(char * pdb_line, struct pe_counter * p_ct)
{
	if (!strncmp("TER", pdb_line, 3)) {
		/* The end of a chain */
		if (0) print_pdb_ter((struct pe_pdb_ter *) pdb_line);
		if (my_protein.sequence[p_ct->sequence].atom_head
		    == p_ct->atom) {	
			/* DNA or RNA */
			fprintf(STDOUT,
"read_pdb_t_line() : Skip sequence[%d] (id= %c) since it is DNA or RNA. \n",
				p_ct->sequence,
				my_protein.sequence[p_ct->sequence].id);
			my_protein.sequence[p_ct->sequence].atom_head = 0;
			my_protein.sequence[p_ct->sequence].atom_tail = 0;
		} else {
			/* Protein */
			my_protein.sequence[p_ct->sequence].atom_tail
				= p_ct->atom - 1;
		}
		p_ct->chain++;
		p_ct->sequence++;
		
		my_protein.sequence[p_ct->sequence].atom_head = p_ct->atom;
		my_protein.sequence[p_ct->sequence].mod_num = p_ct->model;
		my_protein.sequence[p_ct->sequence].id
			= my_protein.chain[p_ct->chain].id;
	}
	return;
}

void read_pdb_general(char * pdb_line, struct pe_counter * p_ct)
{
	switch (*pdb_line) {
		case ('E'):
			read_pdb_e_line(pdb_line, p_ct);
			break;
		case ('M'):
			read_pdb_m_line(pdb_line, p_ct);
			break;
		case ('T'):
			read_pdb_t_line(pdb_line, p_ct);
			break;
	}
	return;
}

void setup_my_protein(struct pe_counter * p_ct)
{
	p_ct->model = 1;
	p_ct->chain = 1;
	p_ct->sequence = 1;
	p_ct->atom = 1;
	my_protein.model[1].seq_head = 1;
	my_protein.sequence[1].atom_head = 1;
	my_protein.sequence[1].id = my_protein.chain[1].id;
	my_protein.sequence[1].mod_num = 1;
	return;
}

void check_pdb_data_format(void)
{
	int i;
	
	if (my_protein.atm_num != 3 * my_protein.res_num) {
		/* Some residues miss N and/or C' atoms. */
		fprintf(STDERR, "\n");
		fprintf(STDERR,
"check_pdb_data_format() : format ERROR (atom num=%d != 3*res num=%d)\n",
			my_protein.atm_num, 3 * my_protein.res_num);
	}
	
	for (i = my_protein.seq_num; i > 0; i--) {
		if (my_protein.sequence[i].atom_tail != 0)
			break;
	}
	if (my_protein.atm_num != my_protein.sequence[i].atom_tail) {
		fprintf(STDERR,
"check_pdb_data_format() : format ERROR (last atom (%d) \n",
			my_protein.atm_num);
		fprintf(STDERR,
"check_pdb_data_format() :               != tail (%d) of the last seq (%d))\n",
			my_protein.sequence[i].atom_tail, i);
		fprintf(STDERR, "\n");
		my_protein.atm_num = my_protein.sequence[i].atom_tail;
	}
	return;
}

/* Fill my_protein.chain[] */
void setup_chains_of_my_protein(void)
{
	int mol_ct, chn_ct;
	
	for (chn_ct = 1; chn_ct <= my_protein.chn_num; chn_ct++) {
		my_protein.chain[chn_ct].atom_head
			= my_protein.sequence[chn_ct].atom_head;
		my_protein.chain[chn_ct].atom_tail
			= my_protein.sequence[chn_ct].atom_tail;
	}
	for (mol_ct = 1; mol_ct <= my_protein.mol_num; mol_ct++) {
		for (chn_ct = my_protein.molecule[mol_ct].chain_head;
		     chn_ct <= my_protein.molecule[mol_ct].chain_tail;
		     chn_ct++) {
			my_protein.chain[chn_ct].mol_id
				= my_protein.molecule[mol_ct].id;
		}
	}
	return;
}

/* Fill molecule-id entry for each sequence */
void setup_molecule_id_of_my_protein(void)
{
	int mol_ct = 1;
	int mdl_ct = 1;
	int chn_ct = 1;
	int i;

	for (i = 1; i <= my_protein.seq_num; i++) {
		if (chn_ct > my_protein.chn_num) {
			chn_ct = 1;
			mdl_ct++;
		}
		if (chn_ct > my_protein.molecule[mol_ct].chain_tail) {
			mol_ct++;
		}
		if (mol_ct > my_protein.mol_num) {
			mol_ct = 1;
			mdl_ct++;
		}
		my_protein.sequence[i].id = my_protein.chain[chn_ct].id;
		my_protein.sequence[i].mol_id = my_protein.molecule[mol_ct].id;
		chn_ct++;
	}
	return;
}

int read_pdb_data(FILE * fp_pdb, struct pe_counter * p_ct)
{
	char pdb_line[LINE_WIDTH + 1];
	
	setup_my_protein(p_ct);

	fseek(fp_pdb, 0L, SEEK_SET);
	while (fgets(pdb_line, LINE_WIDTH + 1, fp_pdb) != NULL) {
		if (*pdb_line == 'A') {
			if (read_pdb_atom(pdb_line, p_ct) == FAIL)
				return FAIL;
		} else {
			read_pdb_general(pdb_line, p_ct);
			/* [FIXME] Only the first NMR model is analyzed. */
			if (p_ct->model > 1) {
				p_ct->model--;
				goto LABEL_OUT;
			}
		}
	}
LABEL_OUT:

	if (my_protein.model[1].seq_tail == 0) {
		my_protein.model[1].seq_tail = my_protein.seq_num;
	}
	my_protein.atm_num = p_ct->atom - 1;
	
	check_pdb_data_format();
	
	setup_chains_of_my_protein();
	
	setup_molecule_id_of_my_protein();
	return SUCCESS;
}

/**
 * read_pdb_file - reads the input pdb file (".ent" file).
 * filename: the name of input pdb file
 *
 * [TODO] Support the HETATM records of PDB files.
 */
int read_pdb_file(char * filename)
{
	FILE * fp_pdb;
	struct pe_counter count;

	fp_pdb = fopen(filename, "r");
	if (fp_pdb == NULL) {
		fprintf(STDERR, "read_pdb_file(): Can't open %s\n", filename);
		return FAIL;
	}

	initialize_protein_data_arrays();
	/*  Scan the pdb file and count the atoms, residues, and so on. */
	if (count_pdb_records(fp_pdb, &count) == FAIL) {
		return FAIL;
	}
	if (0) print_counter(&count);
	if (0) print_pdb_header(&my_pdb_header);
	if (0) print_pdb_master(&my_pdb_master);

	if (allocate_protein_data_arrays(&count) == FAIL)
		return FAIL;
	if (0) {
		fprintf(STDOUT, "\n");
		fprintf(STDOUT,
	"read_pdb_file(): \"my_protein\" has been allocated successfully.\n");
		print_my_protein(&my_protein);
	}
	
	/*  Read atom data from the pdb file. */
	if (read_pdb_data(fp_pdb, &count) == FAIL)
		return FAIL;
	if (0) {
		fprintf(STDOUT, "\n");
		fprintf(STDOUT,
	"read_pdb_file(): \"my_protein\" has been read successfully.\n");
		print_my_protein(&my_protein);
	}
	if (0) {
		fprintf(STDOUT, "\n");
		fprintf(STDOUT, "read_pdb_file(): sequence full dump\n");
		print_part_of_my_sequences(my_protein.atom, 1, my_protein.atm_num);
		fprintf(STDOUT, "read_pdb_file(): end of sequence full dump\n");
	}

	if (fclose(fp_pdb)) {
		fprintf(STDERR, "read_pdb_file(): Can't close input file\n");
		abort();
	}
	return SUCCESS;
}
